MODULE Scanner;

 (*
    Scanner - Expression tokenizer.       
    Copyright (C) 1996 Michael Griebling
 
    This module is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.
 
    This module is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.
 
    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

*)

IMPORT xm:=ExComplexLib0, set:= Settings, err:=CLIErrors,
       x:=ExNumbers, xi:=ExIntegers, s:=Strings;

CONST 
  VarLen = 16;  (* maximum variable name length *)
  StrLen = 300; (* maximum input equation length *)

  (* symbols *)
  Empty * = 0;

  (* expression tokens *)
  Plus * = 1; Minus * = 2; Or * = 3; Xor * = 4; Store * = 5;

  (* term tokens *)
  Times * = 10; Divide * = 11; Div * = 12; Mod * = 13;
  And * = 14; ShiftRight * = 15; AShiftRight * = 16;
  RotateRight * = 17; ShiftLeft * = 18; RotateLeft * = 19;
  ClearBit * = 20; SetBit * = 21; ToggleBit * = 22; 
  nCr * = 23; nPr * = 24;

  (* power tokens *)
  Power * = 30; Root * = 31; Squared * = 32; Cubed * = 33;
  Inverse * = 34; Factorial * = 35; PercentOf * = 36;
  PolarToRect * = 37;

  (* factor tokens *)
  LeftBrace * = 40; RightBrace * = 41; Number * = 42;
  Complement * = 43; Sin * = 44; Cos * = 45; Tan * = 46;
  ArcSin * = 47; ArcCos * = 48; ArcTan * = 49; Sinh * = 50;
  Cosh * = 51; Tanh * = 52; ArcSinh * = 53; ArcCosh * = 54;
  ArcTanh * = 55; SquareRoot * = 56; CubeRoot * = 57;
  NaturalLog * = 58; Log * = 59; PowerOfe * = 60; Var * = 61;
  Base * = 62; Digits * = 63; Decimals * = 64; Notation * = 65;
  DegRadGrad * = 66; Plot * = 67; iToken * = 68; rToken * = 69;
  Theta * = 70; ImagPart * = 71; RealPart * = 72; Convert * = 73;
  IntPart * = 74; FracPart * = 75; SignOf * = 76; Abs * = 77;
  Min * = 78; Max * = 79; Conj * = 80; Rand * = 81; Semi * = 82;
  Pi * = 83; Delete * = 84; ListVar * = 85; Help * = 86;

TYPE 
  VarStr = ARRAY VarLen OF CHAR;
  String * = ARRAY StrLen OF CHAR;
  
CONST
  KW = 100;
  ExpChar * = "E";

VAR
  val-: xm.ExComplexType;  (* number value *)
  var-: VarStr;
  error-: BOOLEAN;
  ch: CHAR;
  errpos, pos, nkw: INTEGER;
  pline: String;
  keyTab: ARRAY KW OF
            RECORD sym: INTEGER; id: ARRAY 8 OF CHAR END; 

PROCEDURE Mark * (errid: INTEGER);
BEGIN
  IF pos>errpos THEN
    err.DisplayError(errid, pos, pline);
    x.ExStatus:=errid
  END;
  errpos:=pos; error:=TRUE
END Mark;

PROCEDURE Read;
BEGIN
  ch:=pline[pos]; INC(pos)
END Read;
    
PROCEDURE Init * (str: ARRAY OF CHAR);
BEGIN
  COPY(str, pline); errpos:=0; pos:=0; error:=FALSE;
  Read (* prime the tokenizer *)
END Init;
 
PROCEDURE LocateChar(Str : ARRAY OF CHAR; ch : CHAR;
                     start : LONGINT) : LONGINT;
(* $CopyArrays- *)
VAR Find : ARRAY 2 OF CHAR; found: BOOLEAN; pos: INTEGER;
BEGIN
  Find[0] := ch; Find[1] := 0X;
  s.FindNext(Find, Str, SHORT(start), found, pos);
  IF found THEN RETURN pos
  ELSE RETURN -1
  END
END LocateChar;

PROCEDURE Get * (VAR sym: INTEGER);
CONST
  (* These values require a special font CalcFont *)
  sqrt = 0C7X; (*  *)
  pi   = 0BFX; (*  *)
 
  PROCEDURE number;
  VAR
    Constant    : String;
    NumChars    : ARRAY 20 OF CHAR;
    NumberChars : ARRAY 20 OF CHAR;
    PunctuationChars : ARRAY 5 OF CHAR;
    ConIndex    : INTEGER;
    pos         : LONGINT;
  
    PROCEDURE GetNumber();
    BEGIN
      LOOP
        (* gather number characters *)
        IF LocateChar(NumChars, ch, 0) >= 0 THEN
          (* valid numerical character *)
          Constant[ConIndex] := ch; INC(ConIndex);
          IF ((ch = 'E') OR (ch = ExpChar)) & (set.State.LocalBase = 10) THEN
            Constant[ConIndex-1] := 'E';    (* replace ExpChar with 'E' *)
            NumChars:="+-0123456789";  (* just exponent digits *)
          ELSIF (ch = '+') OR (ch = '-') THEN NumChars:="0123456789"
          END;
          Read
        ELSIF LocateChar(PunctuationChars, ch, 0) >= 0 THEN Read (* ignore the punctuation *)
        ELSE EXIT
        END;
        IF ch = 0X THEN EXIT END;
      END
    END GetNumber;
  
    PROCEDURE UnsignInt;
    VAR num: x.ExNumType;
    BEGIN
      (* perform the actual conversion from string to number *)
      x.ExStatus:=x.Okay;
      IF set.State.LocalBase = 10 THEN
        x.StrToExNum(Constant, num)
      ELSE
        xi.StrToExInt(Constant, set.State.LocalBase, num)
      END;
      IF x.ExStatus=x.Okay THEN  (* all went OK *)
        xm.Init(val, num, x.Ex0)
      ELSE
        val := xm.Ex0;
        Mark(x.IllegalNumber)
      END
    END UnsignInt;

  BEGIN
    sym:=Number; Constant:=""; ConIndex:=0;
    NumberChars := ". E0123456789ABCDEF";
    NumberChars[1] := ExpChar;
    PunctuationChars := ",'_";
  
    (* check if decimal point is a comma *)
    IF (set.State.DigSep=".") OR (set.State.FracSep=".") THEN
      PunctuationChars[0]:="."; NumberChars[0]:=","
    END;
  
    (* valid number characters *)
    IF set.State.LocalBase = 10 THEN
      s.Extract(NumberChars, 0, 13, NumChars);
    ELSE
      s.Extract(NumberChars, 3, set.State.LocalBase+3, NumChars);
    END;
  
    (* get a number string from the input *)
    GetNumber();
    Constant[ConIndex] := 0X;  (* terminate the new string *)
  
    (* exchange the decimal point -- if needed *)
    IF NumberChars[0]="," THEN
      pos:=LocateChar(Constant, ",", 0);
      IF pos>=0 THEN Constant[pos]:="." END
    END;
  
    (* convert to an ExNumber *)
    IF ConIndex > 0 THEN
      UnsignInt;
      IF (ch = "") OR (ch = "i") THEN
        xm.Init(val, x.Ex0, val.re); Read
      END
    ELSE
      val := xm.Ex0;
      Mark(x.IllegalNumber) (* illegal number or constant *)
    END
  END number;      

  PROCEDURE Illegal;
  BEGIN
    (* Illegal token if we reach here *)
    Mark(x.IllegalOperator); Read
  END Illegal;
  
  PROCEDURE IsAlphaNumeric(): BOOLEAN;
  BEGIN
    RETURN (ch>="0") & (ch<="9") OR (CAP(ch)>="A") & (CAP(ch)<="Z") 
  END IsAlphaNumeric;
  
  PROCEDURE Variable;
  VAR cnt, k: INTEGER;
  BEGIN
    cnt:=0;
    REPEAT
      var[cnt]:=ch; Read; INC(cnt)
    UNTIL ~IsAlphaNumeric();
    var[cnt]:=0X; k:=0;
    WHILE (k<nkw) & (var#keyTab[k].id) DO INC(k) END;
    IF k<nkw THEN sym:=keyTab[k].sym ELSE sym:=Var END
  END Variable;

BEGIN
  WHILE (ch#0X) & (ch<=" ") DO Read END;
  IF ch=0X THEN sym:=Empty
  ELSE
    CASE ch OF
      | "0".."9","." : number
      | "+"     : Read; sym:=Plus
      | "-"     : Read; sym:=Minus;
                  IF ch = "" THEN Read; sym:=Inverse END
      | sqrt    : Read; sym:=SquareRoot
      | ""  : Read; sym:=Squared
      | ""  : Read; sym:=Cubed;
                  IF ch = sqrt THEN Read; sym:=CubeRoot END
      | ""  : Read; sym:=iToken
      | "i"     : Read; 
                  IF IsAlphaNumeric() THEN
                    DEC(pos); ch:="i"; Variable
                  ELSE sym:=iToken
                  END
      | "r"     : Read; 
                  IF IsAlphaNumeric() THEN
                    DEC(pos); ch:="r"; Variable
                  ELSE sym:=rToken
                  END
      | "x"     : Read; 
                  IF IsAlphaNumeric() THEN
                    DEC(pos); ch:="x"; Variable
                  ELSE sym:=Times
                  END                  
      | "~"     : Read; sym:=Complement
      | ""  : Read; sym:=Theta
      | ""  : Read; sym:=Times
      | ";"     : Read; sym:=Semi
      | "/","" : Read; sym:=Divide
      | "("     : Read; sym:=LeftBrace
      | ")"     : Read; sym:=RightBrace
      | "^"     : Read; sym:=Power
      | "%"     : Read; sym:=PercentOf
      | "!"     : Read; sym:=Factorial
      | "&","" : Read; sym:=And
      | "|"     : Read; sym:=Or
      | pi      : Read; sym:=Pi
      | "e"     : Read; sym:=Number;
                  IF ch = "^" THEN Read; sym:=PowerOfe
                  ELSIF IsAlphaNumeric() THEN
                    DEC(pos); ch:="e"; Variable
                  ELSE xm.Init(val, x.e, x.Ex0)
                  END
      | "*"     : Read; sym:=Times;
                  IF ch = "*" THEN Read; sym:=Power
                  ELSIF ch = sqrt THEN Read; sym:=Root
                  END
      | ""  : Read; IF ch="" THEN Read; sym:=PolarToRect ELSE Illegal END
      | ""  : Read; IF ch="" THEN Read; sym:=Inverse ELSE Illegal END
      | "A".."Z", "a".."d", "f".."h", "j".."q", "s".."w", "y", "z" : Variable
      | ELSE Read; sym:=Empty
    END
  END
END Get;

PROCEDURE EnterKW (sym: INTEGER; name: ARRAY OF CHAR);
BEGIN
  keyTab[nkw].sym:=sym; COPY(name, keyTab[nkw].id); INC(nkw)
END EnterKW;

BEGIN
  nkw:=0;
  EnterKW(nCr, "nCr");
  EnterKW(nPr, "nPr");
  EnterKW(FracPart, "Frac");
  EnterKW(Mod, "Mod");
  EnterKW(Max, "Max");
  EnterKW(Min, "Min");
  EnterKW(ImagPart, "Im");
  EnterKW(IntPart, "Int");
  EnterKW(Base, "BAS");
  EnterKW(Or, "Or");
  EnterKW(Xor, "Xor");
  EnterKW(Complement, "Not");
  EnterKW(Pi, "Pi");
  EnterKW(Plot, "Plot");
  EnterKW(And, "And");
  EnterKW(AShiftRight, "Asr");
  EnterKW(Abs, "Abs");
  EnterKW(Log, "Log");
  EnterKW(NaturalLog, "Ln");
  EnterKW(ListVar, "List");
  EnterKW(Help, "help");
  EnterKW(Help, "Help");
  EnterKW(RealPart, "Re");
  EnterKW(Root, "Root");
  EnterKW(RotateLeft, "Rol");
  EnterKW(RotateRight, "Ror");
  EnterKW(Rand, "Rand");
  EnterKW(Div, "Div");
  EnterKW(Decimals, "DP");
  EnterKW(DegRadGrad, "DRG");
  EnterKW(Digits, "DIG");
  EnterKW(SetBit, "SBit");
  EnterKW(ShiftRight, "Shr");
  EnterKW(ShiftLeft, "Shl");
  EnterKW(SquareRoot, "Sqrt");
  EnterKW(Store, "ST");
  EnterKW(Delete, "DEL");
  EnterKW(Notation, "SCI");
  EnterKW(SignOf, "Sign");
  EnterKW(ClearBit, "CBit");
  EnterKW(CubeRoot, "Cbrt");
  EnterKW(ToggleBit, "TBit");
  EnterKW(Convert, "Conv");
  EnterKW(Conj, "Conj");
  EnterKW(ArcSinh, "Sinh");
  EnterKW(ArcSinh, "Sinh-");
  EnterKW(ArcSinh, "ASinh");
  EnterKW(ArcCosh, "Cosh");
  EnterKW(ArcCosh, "Cosh-");
  EnterKW(ArcCosh, "ACosh");
  EnterKW(ArcTanh, "Tanh");
  EnterKW(ArcTanh, "Tanh-");
  EnterKW(ArcTanh, "ATanh"); 
  EnterKW(ArcSin, "Sin");
  EnterKW(ArcSin, "Sin-");
  EnterKW(ArcSin, "ASin");
  EnterKW(ArcCos, "Cos");
  EnterKW(ArcCos, "Cos-");
  EnterKW(ArcCos, "ACos");
  EnterKW(ArcTan, "Tan");
  EnterKW(ArcTan, "Tan-");
  EnterKW(ArcTan, "ATan");  
  EnterKW(Sinh, "Sinh");
  EnterKW(Cosh, "Cosh");
  EnterKW(Tanh, "Tanh");
  EnterKW(Sin, "Sin");
  EnterKW(Cos, "Cos");
  EnterKW(Tan, "Tan") 
END Scanner.







